/*******************************************************************************
 * $Id: ClustalInterface.h,v 1.10 2004/02/27 23:08:55 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#ifndef _ClustalInterface_h_
#define _ClustalInterface_h_

#include "libMems/NumericMatrix.h"
#include "gnFilter.h"
#include "libMems/Match.h"
#include "gn/gnSequence.h"

class ClustalResult;


class ClustalInterface {
public:
	~ClustalInterface(){}
	/**
	 * Returns a reference to a usable ClustalInterface
	 */
	static ClustalInterface& getClustalInterface();
	/**
	 * Attempts to perform a multiple alignment using ClustalW between
	 * <code>r_begin</code> and <code>r_end</code>
	 */
	boolean ClustalAlign( ClustalResult& cr, Match* r_begin, Match* r_end, vector< gnSequence* >& seq_table, boolean flank = true );
	/**
	 * Set the distance matrix to use when computing alignments, writes the guide tree to the location
	 * specified in <code>tree_filename</code>
	 * @param distance_matrix An NxN distance matrix for the sequences
	 * @param tree_filename The output file name for the guide tree
	 */
	void SetDistanceMatrix( NumericMatrix< double >& distance_matrix, string& tree_filename );
	/**
	 * Set the maximum allowed length for a clustal alignment.  Sequences above this length
	 * threshold will be ignored.
	 * @param max_length The maximum clustal length
	 */
	void SetMaxClustalLength( gnSeqI max_length ){ max_clustal_length = max_length; }
	/**
	 * Set the minimum flank size used to anchor alignments on the sequences
	 */
	void SetMinFlankSize( gnSeqI min_flank ){ min_flank_size = min_flank; }
protected:
	void SetDistanceMatrix( NumericMatrix< double >& distance_matrix, string& tree_filename, boolean reread_tree );
	boolean CallClustal( vector< string >& seq_table );
	NumericMatrix< double > distance_matrix;
	gnSeqI max_clustal_length;
	gnSeqI min_flank_size;
	int clustal_score_cutoff;
	bool allocated_aln;
private:
	ClustalInterface( const ClustalInterface& ci ){ *this = ci; }
	ClustalInterface& operator=( const ClustalInterface& ci );
	ClustalInterface();
};


class ClustalResult : public AbstractMatch {
public:
	ClustalResult();
	ClustalResult( uint seq_count, gnSeqI align_length );
	ClustalResult( const ClustalResult& cr );
	~ClustalResult();
	ClustalResult& operator=( const ClustalResult& cr );
	
	ClustalResult* Clone() const { return new ClustalResult( *this ); }
	
	void setAlignment( vector< string >& seq_align, double score );

	double getScore() const;
	string getColumn( gnSeqI colI ) const;
	double getIdentity( uint seqI, uint seqJ ) const;

	vector< string > align_matrix;

	/**
	 * Writes this ClustalResult to the specified output stream (e.g. cout).
	 */
	friend std::ostream& operator<<(std::ostream& os, const ClustalResult& cr); //write to source.

	/**
	 * Reads a ClustalResult from the specified input stream (e.g. cin).
	 */
	friend std::istream& operator>>(std::istream& is, const ClustalResult& cr); //read from source

	gnSeqI align_length;
	double score;

	// Inherited methods from AbstractMatch:
	virtual gnSeqI Length( uint seqI = UINT_MAX ) const; 
	virtual void SetLength( gnSeqI len, uint seqI ) { length[ seqI ] = len; }
	virtual int64 End(uint32 endI) const;
	virtual void Invert();
	virtual int64 GapSize(const AbstractMatch& mhe, uint32 seqI);
	virtual void CropStart(gnSeqI crop_amount);
	virtual void CropEnd(gnSeqI crop_amount);

protected:
	vector< int64 > length;
	NumericMatrix< double > identity;
};

inline
gnSeqI ClustalResult::Length( uint seqI ) const { 
	if( seqI == UINT_MAX )
		return align_length;
	return length[ seqI ]; 
}

inline
int64 ClustalResult::End(uint32 endI) const {
	if( m_start[ endI ] > 0 )
		return m_start[endI] + align_length - 1;
	return m_start[ endI ];
}

inline
void ClustalResult::Invert(){
	const gnFilter* rc_filter = gnFilter::DNAComplementFilter();
	for(uint32 startI = 0; startI < SeqCount(); startI++){
		m_start[startI] = -m_start[startI];
		rc_filter->ReverseFilter( align_matrix[ startI ] );
		
	}
}

inline
void ClustalResult::CropStart(gnSeqI crop_amount){
	if( crop_amount > align_length )
		Throw_gnEx( SeqIndexOutOfBounds() );
	for( uint32 i=0; i < SeqCount(); i++ ){
		align_matrix[ i ] = align_matrix[ i ].substr( crop_amount );
		if( m_start[i] > 0 )
			m_start[i] += crop_amount;
	}
	align_length -= crop_amount;
}

inline
void ClustalResult::CropEnd(gnSeqI crop_amount){
	if( crop_amount > align_length )
		Throw_gnEx( SeqIndexOutOfBounds() );
	align_length -= crop_amount;
	for( uint32 i=0; i < SeqCount(); i++ ){
		align_matrix[ i ] = align_matrix[ i ].substr( 0, align_length );
		if( m_start[i]  < 0 )
			m_start[i] -= crop_amount;
	}
}

inline
int64 ClustalResult::GapSize(const AbstractMatch& mhe, uint32 seqI){
	cerr << "ClustalResult::GapSize not fully implemented\n";
	int64 start_1 = m_start[seqI];
	int64 start_2 = mhe.Start( seqI );

	//check for match existence
	//and for similar match direction
	if((start_2 < 0 && start_1 > 0) ||
		(start_2 > 0 && start_1 < 0) ||
		(start_2 == 0 && start_1 != 0) ||
		(start_2 != 0 && start_1 == 0))
		Throw_gnEx(SeqIndexOutOfBounds());
	
	// this code can probably be simplified
	if(start_2 > 0)
		return start_2 - (start_1 + (int64)align_length);
	if(start_2 < 0)
		return -start_1 - (-start_2 + mhe.Length(seqI));
	return 0;
}


#endif // _ClustalInterface_h_
