/*******************************************************************************
 * $Id: Interval.cpp,v 1.11 2004/02/27 23:08:41 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#include "libMems/Interval.h"
#include "libMems/gnAlignedSequences.h"

int64 Interval::End(uint32 endI) const {
	if( m_start[ endI ] > 0 )
		return m_start[endI] + length[ endI ] - 1;
	return m_start[ endI ];
}

void Interval::Invert(){
	for(uint32 startI = 0; startI < SeqCount(); startI++){
		m_start[startI] = -m_start[startI];
	}
	for( uint matchI = 0; matchI < matches.size(); matchI++ )
		matches[ matchI ]->Invert();

	vector< AbstractMatch* > rev_matches;
	rev_matches.insert( rev_matches.begin(), matches.rbegin(), matches.rend() );
	matches = rev_matches;
}

void Interval::CropStart(gnSeqI crop_amount){
	if( crop_amount > align_length )
		Throw_gnEx( SeqIndexOutOfBounds() );
	for( uint32 i=0; i < SeqCount(); i++ ){
		if( m_start[i] > 0 )
			m_start[i] += crop_amount;
	}
	align_length -= crop_amount;
}

void Interval::CropEnd(gnSeqI crop_amount){
	if( crop_amount > align_length )
		Throw_gnEx( SeqIndexOutOfBounds() );
	align_length -= crop_amount;
	for( uint32 i=0; i < SeqCount(); i++ ){
		if( m_start[i]  < 0 )
			m_start[i] -= crop_amount;
	}
}

int64 Interval::GapSize(const AbstractMatch& mhe, uint32 seqI){
	cerr << "Interval::GapSize not fully implemented\n";
	int64 start_1 = m_start[seqI];
	int64 start_2 = mhe.Start( seqI );

	//check for match existence
	//and for similar match direction
	if((start_2 < 0 && start_1 > 0) ||
		(start_2 > 0 && start_1 < 0) ||
		(start_2 == 0 && start_1 != 0) ||
		(start_2 != 0 && start_1 == 0))
		Throw_gnEx(SeqIndexOutOfBounds());
	
	// this code can probably be simplified
	if(start_2 > 0)
		return start_2 - (start_1 + (int64)align_length);
	if(start_2 < 0)
		return -start_1 - (-start_2 + mhe.Length(seqI));
	return 0;
}

void Interval::CalculateOffset(){

	uint seqI;
	boolean zero_exists = false;
	
	for( seqI = 0; seqI < SeqCount(); seqI++ ){
		m_start[ seqI ] = MEM_NO_MATCH;
		length[ seqI ] = 0;
	}
	// find start in each forward sequence
	uint matchI = 0;
	for(; matchI != matches.size(); matchI++ ){
		zero_exists = false;
		for( seqI = 0; seqI < SeqCount(); seqI++ ){
			if( m_start[ seqI ] == MEM_NO_MATCH )
				m_start[ seqI ] = matches[ matchI ]->Start( seqI );
			if( m_start[ seqI ] < 0 )
				m_start[ seqI ] = 0;
			if( m_start[ seqI ] == MEM_NO_MATCH )
				zero_exists = true;
		}
		if( !zero_exists )
			break;
	}

	// find end in each forward sequence
	for( matchI = matches.size(); matchI > 0; matchI-- ){
		zero_exists = false;
		for( seqI = 0; seqI < SeqCount(); seqI++ ){
			if( matches[ matchI - 1 ]->Start( seqI ) > 0 ){
				if( length[ seqI ] == 0 ){
					length[ seqI ] = matches[ matchI - 1 ]->Start( seqI ) - m_start[ seqI ];
					length[ seqI ] += matches[ matchI - 1 ]->Length( seqI );
				}
			}
			if( length[ seqI ] == 0 )
				zero_exists = true;
		}
		if( !zero_exists )
			break;
	}

	// find start in each reverse sequence
	for( matchI = matches.size(); matchI > 0; matchI-- ){
		zero_exists = false;
		for( seqI = 0; seqI < SeqCount(); seqI++ ){
			if( m_start[ seqI ] == MEM_NO_MATCH )
				m_start[ seqI ] = matches[ matchI - 1 ]->Start( seqI );
			if( m_start[ seqI ] == MEM_NO_MATCH )
				zero_exists = true;
		}
		if( !zero_exists )
			break;
	}

	// find end in each reverse sequence
	for( matchI = 0; matchI != matches.size(); matchI++ ){
		zero_exists = false;
		for( seqI = 0; seqI < SeqCount(); seqI++ ){
			if( matches[ matchI ]->Start( seqI ) < 0 ){
				if( length[ seqI ] == 0 ){
					length[ seqI ] = -(matches[ matchI ]->Start( seqI ) - m_start[ seqI ]);
					length[ seqI ] += matches[ matchI ]->Length( seqI );
				}
			}
			if( length[ seqI ] == 0 )
				zero_exists = true;
		}
		if( !zero_exists )
			break;
	}


	AbstractMatch::CalculateOffset();
}

/**
 * Writes this Interval to the specified output stream (e.g. cout).
 */
std::ostream& operator<<(std::ostream& os, const Interval& cr){
	try{
	for( uint matchI = 0; matchI < cr.matches.size(); matchI++ ){
		const AbstractMatch* m = cr.matches[ matchI ];
		const ClustalResult* clust = dynamic_cast< const ClustalResult* >( m );
		if( clust != NULL )
			os << *clust;
		const Match* match = dynamic_cast< const Match* >( m );
		if( match != NULL )
			os << *match;
		os << endl;
	}
	}catch(...){
		cerr << "Exceptional handler\n";
	}
	return os;
}
/*
void Interval::EliminateOverlaps(){
	// prune off all overlaps in adjacent matches
	for( uint matchI = 1; matchI < matches.size(); matchI++ ){
		const Match* prev_match = dynamic_cast< const Match* >( matches[ matchI - 1] );
		Match* match = dynamic_cast< Match* >( matches[ matchI ] );
		if( prev_match == NULL || match == NULL )
			continue;
		
		gnSeqI max_overlap = 0;
		for(uint seqI = 0; seqI < match->SeqCount(); seqI++ ){
			// check for overlaps
			int64 diff = match->Start( seqI ) - prev_match->End( seqI ) - 1;
			if( match->Start( seqI ) < 0 )
				diff = (match->Start( seqI ) - match->Length()) - prev_match->Start( seqI );
			if( diff < 0 && -diff > max_overlap ){
				max_overlap = -diff;
			}
		}
		if( max_overlap > 0 ){
			if( max_overlap >= match->Length() ){
				matches.erase( matches.begin() + matchI );
				matchI--;
			}else
				match->CropStart( max_overlap );
		}
	}
		
}
*/

void Interval::GetAlignedSequences( gnAlignedSequences& gnas, const vector< gnSequence* >& seq_table ) const {
	vector< string > aln_strings;
	vector< string > aln_names;
	vector< int64 > seq_pos;
	vector< string > match_sequence;
	uint seqI = 0;
	for( ; seqI < seq_table.size(); seqI++ ){
		seq_pos.push_back( 0 );
		string name;
		if( seq_table[ seqI ]->contigListSize() > 0 )
			name = seq_table[ seqI ]->contigName( 0 );
		aln_names.push_back( name );
		aln_strings.push_back( "" );
		match_sequence.push_back( "" );
	}
	gnas.sequences.clear();
	gnas.names.clear();
	gnas.positions.clear();

	const gnFilter* comp_filter = gnFilter::DNAComplementFilter();

	for( uint matchI = 0; matchI < matches.size(); matchI++ ){
		const AbstractMatch* m = matches[ matchI ];
		const ClustalResult* cr = dynamic_cast< const ClustalResult* >( m );
		const Match* match = dynamic_cast< const Match* >( m );
		for( seqI = 0; seqI < seq_table.size(); seqI++ ){
			int64 startI = m->Start( seqI );
			int64 len = m->Length( seqI );
			match_sequence[ seqI ] = "";
//			if( startI >= 1702015 && seqI == 0 )
//				cerr << "";
			if( seq_pos[ seqI ] == 0 )
				seq_pos[ seqI ] = startI < 0 ? startI - len : startI;

			if( cr != NULL ){
//				if( startI == 0 ){
//					aln_strings[ seqI ].append( cr->Length(), '-' );
//					continue;
//				}
				match_sequence[ seqI ] = cr->align_matrix[ seqI ];
			}

			if( match != NULL ){
				if( startI == 0 ){
					aln_strings[ seqI ].append( len, '-' );
					continue;
				}
				// may need to insert gap characters
				// or adjust startI and len for overlaps
				int64 cur_start = startI < 0 ? startI - len : startI;					
				if( seq_pos[ seqI ] < cur_start ){
					// add seqI's actual sequence
					string un_seq;
					try{
					if( startI < 0 )
						un_seq = seq_table[ seqI ]->ToString( cur_start - seq_pos[ seqI ], -cur_start );
					else
						un_seq = seq_table[ seqI ]->ToString( cur_start - seq_pos[ seqI ], seq_pos[ seqI ] );
					}catch(...){
						cout << "unhandled exception\n";
					}
					
					if( startI < 0 )
						comp_filter->ReverseFilter( un_seq );
					aln_strings[ seqI ] += un_seq;
					
					// add gap to the rest
					for( uint seqJ = 0; seqJ < seq_pos.size(); seqJ++ ){
						if( seqJ != seqI )
							aln_strings[ seqJ ].append( cur_start - seq_pos[ seqI ], '-' );
					}
				}
				if( seq_pos[ seqI ] > cur_start ){
					cerr << "Bad Interval Data.  Alignment will be bad\n";
					if( cur_start > 0 )
						startI += seq_pos[ seqI ] - startI;
					len -= seq_pos[ seqI ] - startI;
				}
				if( len <= 0 )	
					continue;
				try{
				match_sequence[ seqI ] = seq_table[ seqI ]->ToString( len, abs( startI ) );
				}catch(...){
					cerr << "unhandled exception\n";
				}
				if( startI < 0 )
					comp_filter->ReverseFilter( match_sequence[ seqI ] );
			}

			if( startI > 0 )
				seq_pos[ seqI ] = startI + len;
			else if( startI < 0 )
				seq_pos[ seqI ] = startI;
		}
		// now that all seqs have had a chance to add gaps, add the aligned sequence
		for( seqI = 0; seqI < seq_table.size(); seqI++ )
			aln_strings[ seqI ] += match_sequence[ seqI ];
		int64 length = aln_strings[ 0 ].length();
		for( seqI = 0; seqI < seq_table.size(); seqI++ )
			if( aln_strings[ seqI ].length() != length )
				cerr << "bad alignment\n";
	}
	
	gnas.positions = vector< int64 >( seq_table.size() );
	for( seqI = 0; seqI < seq_table.size(); seqI++ ){
		gnas.addSequence( aln_strings[ seqI ], aln_names[ seqI ] );
		gnas.positions[ seqI ] = m_start[ seqI ];
	}
}
