/*******************************************************************************
 * $Id: Match.h,v 1.9 2004/02/27 23:08:55 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#ifndef _Match_h_
#define _Match_h_

#include "gnClone.h"
#include <iostream>
#include <vector>
#include <set>
#include <valarray>
#include "libMems/AbstractMatch.h"


/**
 * The Match class stores the location of an <b>equal size</b> matching region
 * between several sequences.  There are numerous functions in this
 * class which can be used to compare and manipulate this match.
 */

class Match : public AbstractMatch {
public:
	enum MemType{
		seed,
		extended
	};

public:
	Match();
	/**
	 * Creates a new Match.
	 * @param seq_count The total number of sequences in the alignment
	 * @param mersize The size of the mers used in the sorted mer lists.
	 * @param m_type The type of mem to create, can either be a seed or already extended.
	 * @see MemType
	 */
	Match( const uint32 seq_count, const gnSeqI mersize, const MemType m_type = seed );
	Match(const Match& mhe);
	Match& operator=(const Match& mhe);
	Match* Clone() const;
	~Match();
	
	/** comparison operator, compares two matches to see if they are the same */
	boolean operator==(const Match& mhe) const;

	/** Returns the length of this match */
	gnSeqI Length( uint seqI = 0 ) const{return m_length;}

	/** Sets the length of this match to @param len */
	void SetLength(gnSeqI len, uint seqI = 0){m_length = len;}

	/** Returns the last coordinate of this match in sequence @param endI */
	int64 End(uint32 endI) const;
	
	/** @return true if this match has already been extended */
	boolean Extended() const{return m_extended;}
	/** Sets this match to be extended if the value passed in "extended" is true */
	void SetExtended(boolean extended){m_extended = extended;}
	/** @return the mer size of the sorted mer lists used to find this match */
	uint32 MerSize() const{return m_mersize;}
	//Match count
	
	/**
	 * Links a subset match to this mem and this mem to the subset.
	 * @param subset The subset match which this mem overlaps
	 */
	virtual void LinkSubset( Match* subset );
	/** Returns a copy of set of the matches which overlap this match */
	set<Match*> GetSubsets() const { return m_subsets; }
	/** Returns a copy of set of the matches which this mem overlaps */
	set<Match*> GetSupersets() const { return m_supersets; }
	
	/** Returns a reference to the set of matches which overlap this match */
	set<Match*>& Subsets() { return m_subsets; }
	/** Returns a reference to the set of matches which this match overlaps */
	set<Match*>& Supersets() { return m_supersets; }

	/**
	 * Inverts the coordinates of this mem.
	 */
	virtual void Invert();
		
	/**
	 * Return the difference between the end of this mem and the start
	 * of mhe in the sequence specified by seqI
	 * @param mhe The mem to compare to.
	 * @param seqI The index of the sequence to compare
	 * @return The gap size.
	 */
	virtual int64 GapSize(const AbstractMatch& mhe, uint32 seqI);
	
	//warning:  none of the following do bounds checking.
	virtual void Move( int64 distance );
	virtual void CropStart(gnSeqI crop_amount);
	virtual void CropEnd(gnSeqI crop_amount);
	virtual void ExtendStart(gnSeqI extend_amount);
	virtual void ExtendEnd(gnSeqI extend_amount);
	//comparison functions

	/**
	 *	Will return true if this mem contains mhe
	 *  Containment implies that a mem has a length >= the contained
	 *  mem, it has coordinates in every genome the contained mem has,
	 *  the difference in start positions in each genome is the same.
	 * @param mhe The mem to check for containment.
	 * @return True if this mem contains mhe.
	 */
	boolean Contains(const Match& mhe) const;

	/**
	 *  Will return true if this mem intersects mhe
	 *  Intersection implies that in sequences which both matches have 
	 *  coordinates the difference in start positions must always be the
	 *  same and the range of the coordinates overlap.  Each match may have
	 *  coordinates in sequences which the other does not.
	 * @param mhe The Match to check for intersection.
	 * @return True if this Match intersects with mhe.
	 */
	virtual boolean Intersects(const Match& mhe) const;
	
	/**
	 *  Will return true if this mem evenly overlaps mhe
	 *  An even overlap implies that for each genome which this mem has 
	 *  coordinates, the difference in start positions must always be the
	 *  same and this mem's coordinates completely contain mhe's in those
	 *  genomes.
	 *  mhe may have coordinates in genomes which this mem does not.
	 * @param mhe The mem to check for intersection.
	 * @return True if this mem intersects with mhe.
	 */
	virtual boolean EvenlyOverlaps(const Match& mhe) const;
	
	/**
	 * Writes the location of this match to the specified output stream (e.g. cout).
	 */
	friend std::ostream& operator<<(std::ostream& os, const Match& mhe); //write to source.

	virtual boolean GetUniqueStart(const Match& mhe, Match& unique_mhe) const;
	virtual boolean GetUniqueEnd(const Match& mhe, Match& unique_mhe) const;

	/** Ensures all linked supersets are really supersets and are correctly linked */
	virtual void UpdateSupersets();
	/** Unlinks this mem from any supersets and subsets it may be linked to. */
	virtual void UnlinkSelf();

	virtual void DetailCompare(const Match& mhe, int& offset, boolean& aligned, int64& diff);
	static boolean offset_lessthan(const Match& a, const Match& b);
	static boolean start_lessthan_ptr(const Match* a, const Match* b);
	static boolean start_lessthan(const Match& a, const Match& b);
	static boolean strict_start_lessthan_ptr(const Match* a, const Match* b);
	/** compare the end of a to the start of b 
	 */
	static int64 end_to_start_compare(const Match& a, const Match& b);
	static int64 start_compare(const Match& a, const Match& b);
	static uint32 seq_compare_start;

	static boolean end_lessthan(const Match& a, const Match& b);
	//see if a is within tolerable range of b
	static int fuzzy_align(const Match& a, const Match& b, uint32 tolerance);


	virtual void AddSubset( Match* overlapper );
	virtual void AddSuperset( Match* underlapperw );
protected:
	void HomogenizeParity(vector<int64>& start1, vector<int64>& start2, const uint32 startI) const;

	gnSeqI m_length;

	virtual void UnlinkSubset( Match* subset );
	virtual void UnlinkSuperset( Match* superset );
	set<Match*> m_subsets;
	set<Match*> m_supersets;

private:

	boolean m_extended;
	gnSeqI m_mersize;

};


//inline
//void Match::SetStart(uint32 seqI, int64 startI){
//	m_start[seqI] = startI;
//}

std::ostream& operator<<(std::ostream& os, const Match& mhe); //write to source.


inline
boolean Match::start_lessthan(const Match& a, const Match& b){
	return start_lessthan_ptr(&a, &b);
}


class MheCompare {
public:
	boolean operator()(const Match* a, const Match* b) const{
		int64 a_matchnum = a->MatchNumber();
		int64 b_matchnum = b->MatchNumber();
		if( a_matchnum < b_matchnum ){
			return true;
		}else if( a_matchnum == b_matchnum ){
			//offsets are the same, check for containment...
			if(a->Contains(*b) || b->Contains(*a)){
				return false;
			}else
				return Match::strict_start_lessthan_ptr(a, b);
		}
		return false;
	}
};
/*
class MemStartComparator {
public:
	MemStartComparator( unsigned seq = 0 ){
		m_seq = seq;
	}
	MemStartComparator( MemStartComparator& msc ){
		m_seq = msc.m_seq;
	}
	// TODO??  make this do a wraparound comparison if all is equal?
	boolean operator()(const Match* a, const Match* b) const{
		int32 start_diff = max( a->FirstStart(), m_seq ) - max( b->FirstStart(), m_seq );
		if(start_diff == 0){
			uint32 m_count = a->SeqCount();
			m_count = m_count <= b->SeqCount() ? m_count : b->SeqCount();
			for(uint32 seqI = m_seq; seqI < m_count; seqI++){
				int64 a_start = a->Start( seqI ), b_start = b->Start( seqI );
				if(a_start < 0)
					a_start = -a_start + a->Length();
				if(b_start < 0)
					b_start = -b_start + b->Length();
				int64 diff = a_start - b_start;
				if(a_start == MEM_NO_MATCH || b_start == MEM_NO_MATCH)
					continue;
				else if(diff == 0)
					continue;
				else
					return diff < 0;
			}
		}
		return start_diff < 0;
	}
private:
	unsigned m_seq;
};
*/
/*
class MheCompare {
public:
	boolean operator()(const Match* a, const Match* b) const{
		int64 a_matchnum = a->MatchNumber();
		int64 b_matchnum = b->MatchNumber();
		if( a_matchnum < b_matchnum ){
			return true;
		}else if( a_matchnum == b_matchnum ){
			//offsets are the same, check for containment...
			if(a->Contains(*b) || b->Contains(*a)){
				return false;
			}else
				return Match::strict_start_lessthan_ptr(a, b);
		}
		return false;
	}
};
*/

#endif // _Match_h_
